
import os
import signal
import subprocess
import argparse
import webbrowser 
import time
import urllib 
import getpass
import glob
import requests

init_facts = [] #fact in initiation rules

# copy contents of both files, and generate a good tamarin file
def init(file1, file2, multi):
	f1 = open(file1, 'r') # first protocol
	f2 = open(file2, 'r') # second protocol
	mul = open(multi, 'w') # output file
	mul.write("theory multi" + os.path.splitext(os.path.basename(file1))[0] + 
		"And" + os.path.splitext(os.path.basename(file2))[0] + "\nbegin \n \n")
	copy_builtins(f1, f2, mul, 'builtins:')
	copy_builtins(f1, f2, mul, 'functions:')
	copy_builtins(f1, f2, mul, 'equations:')
	mul.write("// setup \n") # begin of initialisation rules 
	copy_initiation_rules(f1, f2, mul)
	mul.write("// begin \n") # begin of other rules
	copy_rules(f1, f2, mul)
	copy_deconstruction(f1, f2, mul)
	mul.write("\n")
	copy_restriction(f1, f2, mul)
	mul.write("\nend")
	f1.close()
	f2.close()
	mul.close()
	return

# supress "," when needed, and take only the name of a function
def clean_word(str):
	if str == ",":
		str = ""
	else:
		if str[len(str) - 1] == ",":
			str = str[:len(str) - 1]
		if str[0] == ",":
			str = str[1:len(str)]
	str2 = ""
	# get the name only, if str is a function
	for i in range(len(str)):
		if str[i] == '(':
			break
		str2 = str2 + str[i]
	return str2

# copy builtins of both files
def copy_builtins(f1, f2, mul, type):
	f1.seek(0)
	f2.seek(0)
	for line1 in f1:
		if line1[:len(type)] == type:
			mul.write(line1[:len(line1) - 1])
			break
	if (line1 and not(line1[:len(type)] == type)):
		for line2 in f2:
			if line2[:len(type)] == type:
				mul.write(line2)
				return
	l1 = line1.split()
	for i in range(len(l1)):
		l1[i] = clean_word(l1[i])
	for line2 in f2:
		if line2[:len(type)] == type:
			l2 = line2.split()
			for i in range(len(l2)):
				str = l2[i]
				str = clean_word(str)
				if not(str in l1):
					mul.write(", " + str)
			break
	mul.write("\n")
	return

# copy the initialisation rule only once
def copy_initiation_rules(f1, f2, mul):
	f1.seek(0)
	f2.seek(0)
	for line1 in f1:
		if line1[:8] == '// setup': # begin of initialisation rules
			break
	for line1 in f1:
		if line1[:8] == '// begin':
			break
		for fact in line1.split():
			if fact[0].isupper() or fact[0] == '!':
				fact = clean_word(fact)
				init_facts.append(fact)
		mul.write(line1)
	return

# copy other rule of both protocols
def copy_rules(f1, f2, mul):
	liste = []
	f1.seek(0)
	f2.seek(0)
	boolean = False
	for line1 in f1:
		if line1[:8] == '// begin':
			break
	for line1 in f1:
		if line1[:5] == 'lemma' or line1[:11] == 'restriction' or line1[:6] == '#ifdef':
			break
		mul.write(line1)
	for line2 in f2:
		if line2[:8] == '// begin':
			break
	for line2 in f2:
		if line2[:5] == 'lemma' or line2[:11] == 'restriction' or line2[:6] == '#ifdef':
			break
		# put an 'S' before fact and name of rule to avoid conflict (weak)
		for fact in line2.split():
			fact2 = clean_word(fact)
			if boolean:
				mul.write("S" + fact)
				boolean = False
			# but not for facts in initialisation rule, and facts built in to tamarin
			elif not(fact2 in init_facts) and (fact[0].isupper() or fact[0] == '!') and fact2 != "Fr" and fact2 != "In" and fact2 != "Out" :
				mul.write("S" + fact)
			else:
				mul.write(fact)
			if fact == "rule":
				boolean = True
			mul.write(" ")
		mul.write("\n")
	return

# copy deconstruction lemmas, if they exist
def copy_deconstruction(f1, f2, mul):
	f1.seek(0)
	f2.seek(0)
	boolean1 = False
	boolean2 = False
	for line1 in f1:
		if line1[:21] == "lemma deconstructions":
			boolean1 = True
		if boolean1:
			if line1[len(line1)-2] == "\"":
				mul.write(line1[:len(line1)-2] + "\n")
				break
			mul.write(line1)
	for line2 in f2:
		if line2[:21] == "lemma deconstructions":
			boolean2 = True
			break
	if not boolean2:
		mul.write("\" \n")
		return
	mul.write("&")
	first_line = True
	boolean = False
	for line2 in f2:
		if first_line:
			for fact in line2.split():
				if fact[:1] == "\"":
					mul.write(fact[1:])
				else:
					# put an 'S' before facts to avoid conflicts
					if fact[:4] == "Out_"  or fact[:3] == "In_":
						mul.write(" S" + fact)
					else: 
						mul.write(" " + fact)
			mul.write("\n")
			if line2[len(line2)-2] == "\"":
				break
		else:
			for fact in line2.split():
				# put an 'S' before facts to avoid conflicts
				if fact[:4] == "Out_" or fact[:3] == "In_":
					mul.write(" S" + fact)
				else: 
					mul.write(" " + fact)
			mul.write("\n")
			if line2[len(line2)-2] == "\"":
				break
		first_line = False
	return

# copy all restrictions
def copy_restriction(f1, f2, mul):
	f1.seek(0)
	f2.seek(0)
	boolean = False
	for line1 in f1:
		if line1[:12] == "restriction ":
			mul.write("\n")
			boolean = True
		if boolean:
			mul.write(line1)
		if line1[len(line1) - 2] == "\"":
			boolean = False
	boolean = False
	prev = ""
	for line2 in f2:
		if line2[:12] == "restriction ":
			mul.write("\n")
			boolean = True
		if boolean:
			for fact in line2.split():
				# put an 'S' before facts to avoid conflicts
				if "(" in fact and fact[0] != "(" and fact[:4] != "not(":
					mul.write(" S" + fact)
				elif prev == "restriction":
					mul.write(" S" + fact)	
				else:
					mul.write(" " + fact)
				prev = fact
			mul.write("\n")
		if line2[len(line2) - 2] == "\"":
			boolean = False
	return

# check in the .dot what fact make the lemma is false
def parse_dot(dot, multi):
	fdot = open(dot, 'r')
	In = []
	Out = []
	line_number = 0
	# list all In_dec and Out_dec fact and where they are 
	for line in fdot:
		line_number += 1
		for i in range(len(line)):
			if line[i:i+4] == "SIn_":
				c = 0
				tmp = 0
				while line[i+c-1] != ")" or tmp !=0:
					if line[i+c] == "(":
						tmp += 1
					if line[i+c] == ")":
						tmp -= 1
					c += 1
				In.append((line[i:i+c], line_number))
				i = i + c + 1
			if line[i:i+3] == "In_" and not (line[i-1] == "S"):
				c = 0
				tmp = 0
				while line[i+c-1] != ")" or tmp !=0:
					if line[i+c] == "(":
						tmp += 1
					if line[i+c] == ")":
						tmp -= 1
					c += 1
				In.append((line[i:i+c], line_number))
				i = i + c + 1
			if line[i:i+5] == "SOut_":
				c = 0
				tmp = 0
				while line[i+c-1] != ")" or tmp !=0:
					if line[i+c] == "(":
						tmp += 1
					if line[i+c] == ")":
						tmp -= 1
					c += 1
				Out.append((line[i:i+c], line_number))
				i = i + c + 1	
			if line[i:i+4] == "Out_" and not (line[i-1] == "S"):
				c = 0
				tmp = 0
				while line[i+c-1] != ")" or tmp !=0:
					if line[i+c] == "(":
						tmp += 1
					if line[i+c] == ")":
						tmp -= 1
					c += 1
				Out.append((line[i:i+c], line_number))
				i = i + c + 1
	Validate = False
	# check if all In_dec have a corresponding Out_dec with good arguments
	for i in In:
		Validate = False
		Is_S = False
		Bname = 0
		c = 0
		for j in range(len(i[0])):
			if i[0][0] == "S":
				Is_S = True
			if i[0][j] == "_" and Bname == 0:
				Bname = j + 1
			if i[0][j] == "(" and c != 0:
				c += 1
			if i[0][j] == "(" and c == 0:
				Ename = j
				Barg = j + 1
				c += 1
			if i[0][j] == ")" and c != 0:
				c -= 1	
			if i[0][j] == ")" and c == 0:
				Earg = j
		iname = i[0][Bname:Ename]
		iarg = i[0][Barg:Earg]
		for o in Out:
			Os_S = False
			Bnameo = 0
			for j in range(len(o[0])):
				if o[0][0] == "S":
					Os_S = True
				if o[0][j] == "_" and Bnameo == 0:
					Bnameo = j + 1
				if o[0][j] == "(":
					Enameo = j
					Bargo = j + 1
				if o[0][j] == ")":
					Eargo = j
			oname = o[0][Bnameo:Enameo]
			oarg = o[0][Bargo:Eargo]
			if Is_S == Os_S and iname == oname and iarg == oarg:
				Validate = True
				break
		if not Validate:
			break
	# if all In_dec have their Out_dec, the lemma is true and we succeed
	if Validate or In == []:
		print "done"
		return "OK"
	# build the missng Out_dec
	clause = i[0].replace("~","")
	clause = clause.replace("In_", "Out_")
	clause = clause.replace("\\", "")
	iarg = iarg.replace("(", " ")
	iarg = iarg.replace(")", " ")
	arg = iarg.split()
	for j in range(len(arg)):
		arg[j] = arg[j].replace(",", "")
		arg[j] = arg[j].replace("~", "")
		arg[j] = arg[j].replace("\\", "")
		arg[j] = arg[j].replace("<", "")
		arg[j] = arg[j].replace(">", "")
	fdot.seek(0)
	c = 1
	# find the previous rule to place the Out_dec
	for line in fdot:
		c += 1
		if c == i[1] + 1:
			break 
	pos = ""
	for j in range(len(line)):
		if line[j] == "[" and pos == "":
			pos = line[:j]
		if line[j:j+3] == "In(":
			c = 0
			for k in range(j, 0, -1):
				if line[k] == "<":
					break
			while line[k + c] != ">":
				c += 1
			box = line[k+1:k+c]
			break
	for line in fdot:
		for j in range(len(line)):
			if line[j:j+4+len(pos)+len(box)] == "-> "+pos+":"+box:
				way = line 
				break
	pos = get_pos(fdot, way, arg)
	boolOut = False
	fdot.seek(0)
	Bname = 0
	for line in fdot:
		if line[:len(pos)+1] == pos + "[":
			for j in range(len(pos)+1, len(line)-1):
				if line[j] == ":":
					Bname = j
				if line[j] == "[" and Bname != 0:
					Ename = j
				if line[j:j+4] == "Out(":
					compteur = 0
					Bout = j
					boolOut = True
				if line[j] == "(" and boolOut:
					compteur +=1
				if line[j] == ")" and boolOut:
					compteur -=1
					if compteur == 0:
						OutArg = line[Bout: j]
						break
			name = line[Bname+2:Ename]
			break
	# get the output of the rule 
	OutArg = OutArg.replace("Out(", "")
	OutArg = OutArg.replace(" ) ", " ")
	OutArg = OutArg.replace(",", " ")
	OutArg = OutArg.replace("\\", "")
	OutArg = OutArg.replace("<", " < ")
	OutArg = OutArg.replace(">", " > ")
	OutArg = OutArg.replace(")", " ) ")
	OutArg = OutArg.replace("~", "")
	while OutArg[0] == " " :
		OutArg = OutArg.replace(" ", "", 1)
	if OutArg[0] == "<":
		OutArg = OutArg.replace("<", "", 1)
		while OutArg[len(OutArg)-1] == " " :
			OutArg = OutArg[:len(OutArg)-1]
		if OutArg[len(OutArg)-1] == ">":
			OutArg = OutArg[:len(OutArg)-1]
	OutArg = OutArg.split()
	OutArg = make_list(OutArg)
	fdot.close()
	return (name, clause, OutArg)

# get the position where to place the Out_dec() 
def get_pos(fdot, way, arg):
	boolean = True
	pos = False
	while boolean:
		for j in range(len(way)):
			if way[j] == ":":
				pos = way[:j]
				fdot.seek(0)
				for line in fdot:
					if line[:len(pos) + 1] == pos + "[":
						for i in arg:
							if i not in line:
								return False
				boolean = False
				break
			if way[j] == "\n":
				break	
			if way[j] == " ":
				pos = way[:j]
				fdot.seek(0)
				way = False
				for line in fdot:
					for k in range(len(line)):
						if line[k:k+3+len(pos)] == "-> " + pos:
							way = get_pos(fdot, line, arg) 
							if way != False:
								boolean = False
								break
					if way != False:
						break
				if way != False:
					pos = way
					break
	return pos

# transform arguments of function to list
def make_list(liste):
	tmp = []
	res = []
	type = "none"
	compteur = 0
	for i in liste:
		if compteur != 0:
			if i == ")" or i == ">":
				compteur -= 1
				if compteur == 0:
					res.append(make_list(tmp))
					tmp =[]
			tmp.append(i)
		if i[len(i) - 1] == "(" or i[len(i) - 1] == "<":
			compteur += 1
		if compteur == 0 and i != ")" and i != ">":
			res.append(i)
	return res

# call tamarin and check if the deconstruction lemma is correct
def find_attack(multi):
	user = getpass.getuser()
	cache = "/tmp/tamarin-prover-cache-" + user + "/img/*"
	list_of_files = glob.glob(cache)
	for i in list_of_files:
		os.remove(i)
	tamarin = subprocess.Popen(['~/.local/bin/tamarin-prover interactive .'], shell = True, preexec_fn = os.setsid)
	time.sleep(2) # wait tamarin loads theories
	page = urllib.urlopen('http://127.0.0.1:3001/thy/trace/1/autoprove/idfs/0/proof/deconstructions') 
	strpage = page.read()
	strpage = strpage[13:len(strpage)-2]
	strpage = "http://127.0.0.1:3001"+strpage
	strpage = strpage.replace("/overview/","/graph/")
	r = requests.get(strpage)
	cache = "/tmp/tamarin-prover-cache-" + user + "/img/*.dot"
	list_of_files = glob.glob(cache) # * means all if need specific format then *.csv
	latest_file = max(list_of_files, key=os.path.getctime)
	var = parse_dot(latest_file, multi)
	os.killpg(os.getpgid(tamarin.pid), signal.SIGTERM)
	return var

# place the Outdec in the right rule
def correct_dec(name, clause, OutArg1, multi):
	mul = open(multi, "r+") 
	boolean = False
	c = 0
	se = 0
	compteur = 0
	boolOut = False
	# place the seek in the right rule
	for line in mul:
		if line[:4] == "rule" and boolean:
			break
		if line[:5+len(name)] == "rule " + name:
			boolean = True
		if boolean:
			for i in range(len(line)):
				if line[i:i+4] == "Out(":
					compteur = 0 
					Bout = i
					boolOut = True
				if line[i] == "(" and boolOut:
					compteur +=1
				if line[i] == ")" and boolOut:
					compteur -=1
					if compteur == 0:
						OutArg = line[Bout: i]
						break
	# replace arguments from trace to coreesponding in .spthy file
	OutArg = OutArg.replace("Out(", "")
	OutArg = OutArg.replace(" ) ", " ")
	OutArg = OutArg.replace(",", " ")
	OutArg = OutArg.replace("\\", "")
	OutArg = OutArg.replace("<", " < ")
	OutArg = OutArg.replace(">", " > ")
	OutArg = OutArg.replace(")", " ) ")
	OutArg = OutArg.replace("~", "")
	while OutArg[0] == " " :
		OutArg = OutArg.replace(" ", "", 1)
	if OutArg[0] == "<":
		OutArg = OutArg.replace("<", "", 1)
		while OutArg[len(OutArg)-1] == " " :
			OutArg = OutArg[:len(OutArg)-1]
		if OutArg[len(OutArg)-1] == ">":
			OutArg = OutArg[:len(OutArg)-1]
	OutArg = OutArg.split()	
	OutArg = make_list(OutArg)
	for j in range(len(clause)):
		if clause[j] == "(":
			break
	clause = clause_anti_rewrite(clause, OutArg1)
	anti_rewrite(OutArg1)
	clause = replace_Out_dec(OutArg1, OutArg, j, clause)
	clause = clause.replace("+", "")
	var = clause.split()
	del var[0]
	del var[len(var)-1]
	while'<' in var:
		var.remove('<')
	while ',' in var:
		var.remove(',')
	while '>' in var:
		var.remove('>')
	while '(' in var:
		var.remove('(')
	while ')' in var:
		var.remove(')')
	mul.seek(0)
	boolean = False
	# write the Out_dec() in the file
	for line in mul:
		if line[:4] == "rule" and boolean:
			break
		if line[:5+len(name)] == "rule " + name:
			boolean = True
		if boolean:
			for arg in var:
				for j in range(len(line)):
					if line[j:j+len(arg)] == arg and line[j-1] == "~":
						for i in range(len(clause)):
							if clause[i:i+len(arg)] == arg and clause[i-1] != "~":
								clause = clause[:i] + "~" + clause[i:]
			for i in range(len(line)):
				se += 1
				if line[i] == "[":
					c +=1
					if c == 2:
						tmp = line[i + 1:]
						for line2 in mul:
							tmp  = tmp + line2
						mul.seek(se)
						mul.writelines(" " + clause + "," + tmp)
						mul.flush()
						break
		else:
			se = se + len(line)
	mul.close()

# avoid multiple rewrite on a variable
def clause_anti_rewrite(clause, OutArg):
	clause = clause.replace("(", "( ")
	clause = clause.replace("<", " < ")
	clause = clause.replace(">", " > ")
	clause = clause.replace(",", " , ")
	clause = clause.replace(")", " ) ")
	clause = clause.split()
	for i in range(len(clause)):
		if not (clause[i][len(clause[i]) -1] == "(" or clause[i] == "<" or clause[i] == ">" or clause[i] == "," or clause[i] == ")"):
			clause[i] = "+"+clause[i]+"+"
	clause = " ".join(clause)
	return clause

# avoid multiple rewrite on a variable
def anti_rewrite(liste):
	for i in range(len(liste)):
		if isinstance(liste[i], list):
			anti_rewrite(liste[i])
		else:
			liste[i] = "+" + liste[i] + "+"

# replace arguments of the Out_dec from the .dot to those on the .spthy
def replace_Out_dec(OutArg1, OutArg2, place, clause):
	if len(OutArg2) == len(OutArg1):
		while len(OutArg1) != 0:
			i = min_alpha(OutArg1)
			if isinstance(OutArg1[i], list) and isinstance(OutArg2[i], list) and len(OutArg1[i]) == len(OutArg2[i]):
				clause = replace_Out_dec(OutArg1[i], OutArg2[i], place, clause)
			else:
				clause = clause[:place] + clause[place:].replace(str(OutArg1[i]), str(OutArg2[i]))
			del OutArg1[i]
			del OutArg2[i]
	return clause

# get the min of a string list
def min_alpha(liste):
	min = liste[0]
	indice = 0
	for i in range(len(liste)):
		 if min != alpha(min, liste[i]):
		 	min = alpha(min, liste[i])
		 	indice = i
	return indice

# compare two string
def alpha(str1, str2):
	if len(str1) > len(str2):
            return str1
        if len(str1)==len(str2):
            return min(str1, str2)
        else:
            return str2

# correct the special case that cause loop
def correct_special_case(mul, factIn, factOut, rule1, rule2):
	mul.seek(0) 
	Outmax = factIn.find("(")
	factOut1 = factIn[:Outmax]
	factIn1 = factOut1.replace("Out", "In", 1)
	Inmax = factOut.find("(")
	factIn2 = factOut[:Inmax]
	factOut2 = factIn2.replace("In", "Out", 1)
	boolean = False
	se = 0
	# erase the Out_dec1() and Out_dec2() in both rules that cause loop
	for line in mul:
	 	if line[:4] == "rule":
	 		boolean = False
	 		name = ""
	 		for i in line[4:]:
	 			if i == ":":
	 				break
	 			if i != " ":
	 				name = name + i
	 		if name == rule1 or name == rule2:
	 			boolean = True
	 	if boolean:
	 		for i in range(len(line)):
	 			if line[i:i+len(factOut1)] == factOut1:
	 				mul.seek(se + i)
	 				c = 0
	 				b = True
	 				while b:
	 					mul.write(" ")
	 					mul.flush()
	 					if line[i] == "," and c == 0:
	 						b = False
	 					if line[i] == "(":
	 						c += 1
	 					if line[i] == ")":
	 						c -= 1
	 					i += 1
	 				break
	 			if line[i:i+len(factOut2)] == factOut2:
	 				mul.seek(se + i)
	 				c = 0
	 				b = True
	 				while b:
	 					mul.write(" ")
	 					mul.flush()
	 					if line[i] == "," and c == 0:
	 						b = False
	 					if line[i] == "(":
	 						c += 1
	 					if line[i] == ")":
	 						c -= 1	
	 					i += 1
	 				break
	 	se = se + len(line) 
		mul.seek(se)
	mul.seek(0) 
	se = 0
	# replace all Out_dec1(), Out_dec2() and, In_dec1(), In_dec2() in all file by th shortest of them
	for line in mul:
		for i in range(len(line)):
	 		if line[i:i+len(factOut1)] == factOut1 and len(factOut1) >= len(factOut2):
	 			mul.seek(se + i)
	 			c = 0
	 			while c < len(factOut1):
	 				if c < len(factOut1) - len(factOut2):
	 					mul.write(" ")
	 					mul.flush()
	 					c += 1
	 				else:
	 					mul.write(factOut2)
	 					mul.flush()
	 					c += len(factOut2)	 						
	 			break
	 		if line[i:i+len(factIn1)] == factIn1 and len(factIn1) >= len(factIn2):
	 			mul.seek(se + i)
	 			c = 0
	 			while c < len(factIn1):
	 				if c < len(factIn1) - len(factIn2):
	 					mul.write(" ")
	 					mul.flush()
	 					c += 1
	 				else:
	 					mul.write(factIn2)
	 					mul.flush()
	 					c += len(factIn2)	 						
	 			break	
	 		if line[i:i+len(factOut2)] == factOut2 and len(factOut2) > len(factOut1):
	 			mul.seek(se + i)
	 			c = 0
	 			while c < len(factOut2):
	 				if c < len(factOut2) - len(factOut1):
	 					mul.write(" ")
	 					mul.flush()
	 					c += 1
	 				else:
	 					mul.write(factOut1)
	 					mul.flush()
	 					c += len(factOut1)
	 			break
	 		if line[i:i+len(factIn2)] == factIn2 and len(factIn2) > len(factIn1):
	 			mul.seek(se + i)
	 			c = 0
	 			while c < len(factIn2):
	 				if c < len(factIn2) - len(factIn1):
	 					mul.write(" ")
	 					mul.flush()
	 					c += 1
	 				else:
	 					mul.write(factIn1)
	 					mul.flush()
	 					c += len(factIn1)
	 			break
	 	se = se + len(line) 
		mul.seek(se)				

# check if we have the special case that cause loop
def check_special_case(multi):
	mul = open(multi, "r+")
	j = -1
	res = []
	name = ""
	boolean = False
	Is_OutorIn = False
	Term = False
	# make a list with all In_dec() and Out_dec() by rules
	for line in mul:
		strdec = ""
		if line[:4] == "rule":
			boolean = False
			for i in range(4, len(line)):
				if line[i] == ":":
					break
				if line[i-4:i] == "rule":
					continue
				if line[i] == " ":
					continue
				name = name + line[i]
			res.append([name,[]])
			name = ""
			j += 1
		for i in range(len(line)):
			if line[i:i+3] == "--[":
				boolean = True
			if line[i] == "]":
				boolean = False
				break
			if (line[i:i+3] == "In_" 
				or line[i:i+4] == "SIn_" 
				or line[i:i+4] == "Out_" 
				or line[i:i+5] == "SOut_") and boolean:
				Is_OutorIn = True
				compteur = 0
			if Is_OutorIn:
				if line[i] == "(":
					compteur += 1
				if line[i] == ")":
					compteur -= 1
					if compteur == 0:
						strdec = strdec + line[i]
						Is_OutorIn = False
						Term = True
				strdec = strdec + line[i]
			if strdec != "" and Term:
				strdec = strdec.replace(" ","")
				strdec = strdec.replace("~","")
				res[j][1].append(strdec)
				strdec = ""
				Term = False
	# check in the list if we have the special case
	for rule in res:
		factIn = ""
		factOut = ""
		for fact1 in rule[1]:
			if fact1[:3] == "In_" or fact1[:4] == "SIn_":
				factIn = fact1
				factIn = factIn.replace("In", "Out", 1)
				for fact2 in rule[1]:
					if fact2[:4] == "Out_" or fact2[:5] == "SOut_":
						factOut = fact2
						factOut = factOut.replace("Out", "In", 1)
						for rule2 in res:
							if rule2 != rule:
								if (factIn in rule2[1] and factOut in rule2[1]):
									correct_special_case(mul, factIn, factOut, rule[0], rule2[0])
									break
						SOut = False
	mul.close()

def main():
	os.system("killall tamarin-prover")
	parser = argparse.ArgumentParser()
	parser.add_argument('file1', help='first protocol (.spthy) path')
	parser.add_argument('file2', help='second protocol (.spthy) path')
	args = parser.parse_args()
	init(args.file1, args.file2, multi)
	boolean = True
	while boolean:
		var = find_attack(multi)
		if var == "OK":
			boolean = False
		else:
			correct_dec(var[0], var[1], var[2], multi)
		check_special_case(multi)

### test ###

multi = "multi_test.spthy"

main()
